/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLEDMISC.C                        Date Last Modified: 06-Feb-02 *
*                                                                       *
* Miscellaneous functions for NLED.                                     *
* Functions are:                                                        *
*   atoindex() - converts character string to INDEX data type           *
*   doscall() - interactive call of DOS command                         *
*   find() - finds a given string in data and moves there               *
*   findrepl() - interactive find and, optionally, replace              *
*   funckey() - translates keycode into 0-39, or -1 if not function key *
*   gotoline() - interactive go to given line of data                   *
*   htoc() - converts 2-digit hex code to single byte                   *
*   inpfunckey() - interactive programming of function key              *
*   inputhex() - interactive input of byte using hex code               *
*   setbeep() - interactive setting of attention (beep) level           *
*   setcase() - interactive setting of case sensitivity                 *
*   setmode() - interactive setting of file access mode                 *
*   settabs() - interactive setting of tab_size                         *
*   sstrcmp() - compares fixed number of characters in two arrays       *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
* 20-Mar-95: Linux port                                                 *
* 08-Jul-96: Use correct system header files                            *
* 06-Feb-02: Release 2.5 cleanup (mostly longer string editing)         *
\***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "nled.h"
#include "nledkeys.h"
#include "nledvdef.h"
#include "nledfdef.h"

/***********************************************************************\
* Converts null terminated str[] to INDEX data type, if possible.       *
* Returns the result of the conversion or 0.                            *
\***********************************************************************/
INDEX atoindex(char *str)
{
    INDEX n = 0;
    int sign = 1;

    while (' ' == *str) str++;

    if ('-' == *str) {
        sign = -1;
        str++;
    }

    while (('0' <= *str) && (*str <= '9'))
        n = (10 * n) + *(str++) - '0';

    return(sign * n);
}

/***********************************************************************\
* Interactively allows user to run any DOS command. Returns ERR if DOS  *
* could not be invoked (eg. not enough memory), else returns 0.         *
* 12-May-87: invoke COMMAND if user doesn't specify command.            *
* 13-Feb-92: UNIX command support                                       *
* 20-Mar-95: cover functions for ALL curses routines                    *
* 02-Feb-02: increased length of command from 30 to 100                 *
* 06-Feb-02: use resetterm and fixterm for Win95                        *
\***********************************************************************/
int doscall(void)
{
#if ZORTECH
    static int room = TRUE;
#else
    int room = TRUE;
#endif
    char command[101];
    int len;

    /* NOTE: if DOS call fails once, second try sometimes bombs program */
    if (room) {
        len = 0;
#if UNIX
        if (0 > inputs(command,&len,100," UNIX command:")) 
            return (ESCHIT);
#else
        if (0 > inputs(command,&len,100," DOS command:")) 
            return (ESCHIT);
        if (len == 0)
            /* if they didn't enter anything, call up another DOS */
            strcpy(command, "COMMAND");
        else
#endif
            /* else terminate string with a null */
            command[len] = '\0';
        qd_clear();
        qd_flush();

#if UNIX
        if (command[0]) {
            qd_saveterm();
            qd_resetterm();
            system(command);
            qd_fixterm();
        }
        { /* no if needed in UNIX case, so just use open bracket */
#else
        qd_resetterm();
        if (0 == system(command)) {
            qd_fixterm();
#endif
            dispmess(" Press any key to continue",27,FALSE);
            inputc();
#if UNIX
            qd_clear();
#endif
            setdrawall();
            return (OK);
        }
    }

    room = FALSE;
#if WIN95
    dispmess(" Command failed", -1, FALSE);
#else
    dispmess(" No room to load DOS",-1,FALSE);
#endif
    beep(2);
    setdrawall();
    return (ERR);
}

/***********************************************************************\
* Tries to find 'string[]' (length 'len') in data starting at current   *
* position. If found, it moves to the start of the occurence found and  *
* returns OK, otherwise it returns ERR, staying at the current position *
* 27-Dec-88: unsigned char for Zortech C                                *
* 13-Feb-92: made case sensitivity optional                             *
\***********************************************************************/
int find(char *string, int len)
{
    INDEX row, col, pos;
    int i, c, c1, t, t1;

    row = scr_top + row_no;
    col = scr_left + col_no;
    pos = abs_pos;
    datfindseq(pos);

    while (pos <= len_ds - len) {
        c = datgetseq();
        c1 = (unsigned char)string[0];
        if (!case_sens) {
            c = toupper(c);
            c1 = toupper(c1);
        }
        if (c == c1) {
            for (i = 1; i < len; i++) {
                t = datget(pos + i);
                t1 = (unsigned char)string[i];
                if (!case_sens) {
                    t = toupper(t);
                    t1 = toupper(t1);
                }
                if (t != t1) break;
            }
            if (i == len) {
                dispupd('r',row - scr_top - row_no);
                dispupd('c',-scr_left - col_no);
                dispupd('c',col);
                return(OK);
                break;
            }
        }
        if (LF == c) {
            row++;
            col = 0;
        }
        else    col++;
        pos++;
    }
    return(ERR);
}

/***********************************************************************\
* Interactive control of find ('mode' == 0) or find and replace ('mode' *
* == 1). Returns ERR, OK or ESCHIT depending on the situation.          *
\***********************************************************************/
int findrepl(int mode /* 0->find, 1->find and replace */)
{
    int nogood,c,all = FALSE,drow,dcol,nrow,ncol,i,n,tlen;
    static int len = 0, nlen = 0;
    static char string[31] = {""};
    static char newstring[31] = {""};
    char tstring[31];

    tlen = 0;
    insert(string, len, 0, tstring, &tlen, 30);

    if (ESCHIT == inputs(tstring,&tlen,30," Find:"))
        return(ESCHIT);

    if ((tlen != len) || sstrcmp(string,tstring,len)) {
        len = nlen = 0;
        insert(tstring, tlen, 0, string, &len, 30);
        insert(tstring, tlen, 0, newstring, &nlen, 30);
    }

    if (0 == len) {
        dispmess(" No string to find",-1,FALSE);
        return(ERR);
    }

    drow = 0;
    dcol = 0;
    for (i = 0; i < len; i++)
        if (LF == string[i]) {
            drow++;
            dcol = 0;
        }
        else dcol++;

    if (mode == 1) {
        tlen = 0;
        insert(newstring, nlen, 0, tstring, &tlen, 30);
        if (ESCHIT == inputs(tstring, &tlen, 30, " Replace with:"))
            return(ESCHIT);
        nlen = 0;
        insert(tstring, tlen, 0, newstring, &nlen, 30);
    }

    do {
        dispmess(" Trying to find string...",25,TRUE);
        if (ERR == find(string,len)) {
            dispmess(" String not found",-1,FALSE);
            return(ERR);
        }

        if (mode == 0) {
            if (drow) {
                dispupd('r', drow);
                dispupd('c', -scr_left - col_no);
            }
            dispupd('c',dcol);
            return (OK);
        }

        do {
            display();
            if (all) {
                qd_flush();
                c = 'R';
            }
            else {
                dispmess(" Replace, Skip or All? (R, S, A or <Esc>)",-1,FALSE);
                c = inputc();
                c = toupper(c);
            }
            switch (c) {
                case 'A':
                    all = TRUE;
                case 'R':
                    datdeln(len,abs_pos);
                    n = datinss(newstring,nlen,abs_pos);
                    len_ds += n - len;
                    nrow = 0;
                    ncol = 0;
                    for (i = 0; i < n; i++)
                        if (LF == newstring[i]) {
                            nrow++;
                            ncol = 0;
                        }
                        else ncol++;
                    nol_ds += (nrow - drow);
                    dispcalc(row_no);
                    blockupd(abs_pos, n - len);
                    setdraw();
                    if (drow || nrow) draw_erow = num_rows - 2;
                    if (nrow) {
                        dispupd('r', nrow);
                        dispupd('c', -scr_left - col_no);
                    }
                    dispupd('c', ncol);
                    display();
                    qd_flush();
                    edit_flag = TRUE;
                    nogood = FALSE;
                    break;

                case 'S':
                    if (drow) {
                        dispupd('r', drow);
                        dispupd('c', -scr_left - col_no);
                    }
                    dispupd('c',dcol);
                    nogood = FALSE;
                    break;

                case ESC:
                    return(ESCHIT);
                    break;

                default:
                    beep(1);
                    nogood = TRUE;
                    break;
            }
        } while (nogood);
    } while (TRUE);
}

/***********************************************************************\
* Determines if 'c' represents a function key code and returns function *
* key number (0 - 39) if it does, -1 if it doesn't.                     *
* 13-Feb-92: UNIX function keys                                         *
\***********************************************************************/
int funckey(int c)
{
#if UNIX
    if (c < FMIN || c > FMAX) return (-1);
    return (c - FMIN);
#else
    if (!(c & 0xFF00)) return (-1);

    c &= 0xFF;

    if ((c < 0x3B) || (c > 0x71)) return (-1);

    /* F1 to F10: return 0 - 9 (F1 is 0x3B) */
    if (c < 0x45) return (c - 0x3B);

    /* 0x45 - 0x53 are not function keys */
    if (c < 0x54) return (-1); 

    /* Shift/F1 - Shift/F10 (0x54 - 0x5D): return 10 - 19 */
    /* Ctl/F1   - Ctl/F10   (0x5E - 0x67): return 20 - 29 */
    /* Alt/F1   - Alt/F10   (0x68 - 0x71): return 30 - 39 */
    return (c + (10 - 0x54));
#endif
}

/***********************************************************************\
* Interactively allows user to go to a specific line in the data.       *
* 02-Feb-02: allow up to 10 digits instead of 6                         *
\***********************************************************************/
int gotoline(void)
{
    int n,len;
    INDEX line;
    char str[11];

    len = 0;

    n = inputs(str, &len, 10, " Go to line:");
    if ((ESCHIT == n) || (0 == n))
        return(ESCHIT);
    str[len] = '\0';

    if (0 >= (line = atoindex(str))) {
        dispmess(" Invalid line number specified", -1, FALSE);
        return(ERR);
    }

    if (line > nol_ds) line = nol_ds;
    dispupd('r',line - row_no - scr_top - 1);
    offedge();
    return(OK);
}

/***********************************************************************\
* Converts 2 hex digit 'string[]' to character. Returns character code  *
* or -1 if conversion could not be done.                                *
* 14-Feb-92: include lower case a-f                                     *
\***********************************************************************/
int htoc(char *string)
{
    int retcode = 0, i;

    for (i = 0; i < strlen(string); i++) {
        retcode <<= 4;
        if ((string[i] >= '0') && (string[i] <= '9'))
            retcode |= string[i] - '0';
        else if ((string[i] >= 'A') && (string[i] <= 'F')) 
            retcode |= string[i] - 'A' + 10;
        else if ((string[i] >= 'a') && (string[i] <= 'f')) 
            retcode |= string[i] - 'a' + 10;
        else
            return (-1);
    }
    return(retcode);
}

/***********************************************************************\
* Interactively lets user enter function key sequence. Returns OK if    *
* sequence was entered, or ESCHIT if user cancelled.                    *
* 01-APR-87: Removed unused variable "c"                                *
* 13-Feb-92: UNIX function keys                                         *
* 03-Apr-95: Fixed F10 bug in UNIX function keys                        *
\***********************************************************************/
int inpfunckey(void)
{
    int bad,fk,n,i;
    char keyname[15],message[51];

    do {
        dispmess(" Press the function key to program   (or <Esc>)",35,FALSE);
        fk = inputc();
        if (fk == ESC) return (ESCHIT);
        if ((fk = funckey(fk)) < 0) {
            beep(1);
            bad = TRUE;
        }
        else bad = FALSE;
    } while (bad);

#if UNIX
    if (fk >= 9) {
        strcpy(keyname,"<F  >");
        keyname[2] = '0' + ((fk+1)/10);
        keyname[3] = '0' + ((fk+1)%10);
    }
    else {
        strcpy(keyname,"<F >");
        keyname[2] = '1' + fk%10;
    }
#else
    if (fk >= 10) {
        if (fk >= 20) {
            if (fk >= 30) strcpy(keyname,"<Alt/F >");
            else strcpy(keyname,"<Ctl/F >");
        }
        else strcpy(keyname,"<Shft/F >");
    }
    else strcpy(keyname,"<F >");

    if ((1+fk)%10)
        keyname[strlen(keyname) - 2] = '1'+ (fk%10);
    else {
        keyname[strlen(keyname) - 2] = '\0';
        strcat(keyname,"10>");
    }
#endif

    strcpy(message," How many keys for ");
    strcat(message,keyname);
    strcat(message,"?   (0-9 or <Esc>)");
    do {
        dispmess(message,strlen(message) - 16,FALSE);
        n = inputc();
        if (n == ESC) return(ESCHIT);
        if ((n < '0') || (n > '9')) {
            beep(1);
            bad = TRUE;
        }
        else bad = FALSE;
    } while (bad);

    n = n - '0';
    func_mac[fk][0] = n;
    strcpy(message," Programming ");
    strcat(message,keyname);
    strcat(message,": press key  ");
    for (i = 1; i <= n; i++) {
        message[strlen(message) - 1] = i + '0';
        dispmess(message, strlen(message) + 1,FALSE);
        func_mac[fk][i] = inputc();
    }
    return(OK);
}

/***********************************************************************\
* Allows user to enter 2 digit hex code on bottom line of screen.       *
* Returns character entered or ESCHIT if user cancelled.                *
* 27-APR-87: use qd_puts instead of qd_putc                             *
\***********************************************************************/
int inputhex(void)
{
    char str[3];
    int c, len, pos;

    str[0] = str[1] = str[2] = '\0';
    qd_move(num_rows-1, 73);
    len = pos = 0;

    while (TRUE) {
        qd_move(num_rows-1, 73);
        qd_puts(2,str);
        qd_move(num_rows-1, 73+pos);

        status_flag = FALSE;
        c = inputc();
        status_flag = TRUE;
        c = toupper(c);
        switch (c) {
        case ESC:
            return (ESCHIT);

        case CR:
        case LF:
            if (0 == len) return (ESCHIT);
            else return(htoc(str));
            break;

        case BS:
            if (pos == 0) break;
            pos--;
        case DEL:
            if (pos == 0) str[0] = str[1];
            if (pos <= 1) str[1] = '\0';
            break;

        case LEFT:
            if (pos > 0) pos--;
            break;

        case RGHT:
            if (pos < len) pos++;
            break;

        default:
            if ((pos < 2) &&
             ((('0' <= c) && (c <= '9')) || (('A' <= c) && (c <= 'F')))) {
                if (ins_mode && (pos == 0)) str[1] = str[0];
                str[pos] = c;
                pos++;
            }
            else beep(1);
            break;
        }
        len = strlen(str);
    }
}

/***********************************************************************\
* Allows user to interactively set the severity at which the console    *
* beeps, from 0 (never beep) to 2 (beep on anything suspicious).        *
\***********************************************************************/
int setbeep(void)
{
    int c;

    while (TRUE) {
        dispmess(" Attention level?   (0, 1 or 2)",18,FALSE);
        c = inputc();
        switch (c) {
        case ESC:
        case LF:
        case CR:
            return(ESCHIT);
            break;

        case '0':
            dispmess(" Set not to beep on errors", -1, FALSE);
            beep_level = 0;
            return(OK);
            break;

        case '1':
            dispmess(" Set to beep only on severe errors", -1, FALSE);
            beep_level = 1;
            return(OK);
            break;

        case '2':
            dispmess(" Set to beep on input errors", -1, FALSE);
            beep_level = 2;
            return(OK);
            break;

        default:
            beep(1);
            break;
        }
    }
}

/***********************************************************************\
* Allows user to interactively set whether or not search (find or find/ *
* replace) are case sensitive.                                          *
* 13-Feb-92: new routine                                                *
\***********************************************************************/
int setcase(void)
{
    int c;

    while (TRUE) {
        dispmess(" Case sensitive searches?   [Y or N]",26,FALSE);
        c = inputc();
        c = toupper(c);
        switch (c) {
        case ESC:
        case LF:
        case CR:
            return(ESCHIT);
            break;

        case 'Y':
            dispmess(" Case sensitivity activated", -1, FALSE);
            case_sens = TRUE;
            return(OK);
            break;

        case 'N':
            dispmess(" Case sensitivity deactivated", -1, FALSE);
            case_sens = FALSE;
            return(OK);
            break;

        default:
            beep(1);
            break;
        }
    }
}

/***********************************************************************\
* Interactively allows user to set file access mode.                    *
\***********************************************************************/
int setmode(void)
{
    int c;

    while (TRUE) {
        dispmess(" File mode?   (B[inary], T[ext] or Z[EOF text])", 
         12, FALSE);
        c = inputc();
        c = toupper(c);
        switch (c) {
        case ESC:
        case CR:
        case LF:
            return(ESCHIT);
            break;

        case 'T':
            dispmess(" File mode set to Text",-1,FALSE);
            text_mode = TEXT;
            return(OK);
            break;

        case 'Z':
            dispmess(" File mode set to ^Z-end-of-file Text",-1,FALSE);
            text_mode = TEXTZ;
            return(OK);
            break;

        case 'B':
            dispmess(" File mode set to Binary",-1,FALSE);
            text_mode = FALSE;
            return(OK);
            break;

        default:
            beep(1);
            break;
        }
    }
}

/***********************************************************************\
* Interactively allows user to set number of spaces per tab stop        *
\***********************************************************************/
int settabs(void)
{
    int c;

    while (TRUE) {
        dispmess(" How many spaces for each tab?   (1-F)",31,FALSE);
        c = inputc();
        c = toupper(c);
        if (('1' <= c) && (c <= '9')) {
            tab_size = c - '0';
            break;
        }
        if (('A' <= c) && (c <= 'F')) {
            tab_size = 10 + c - 'A';
            break;
        }
        if ((c == LF) || (c == CR) || (c == ESC))
            return(OK);
        beep(1);
    }
    dispmess(" Tab size reset",-1,FALSE);
    return(OK);
}

/***********************************************************************\
* Compares up to 'len' characters of 'string1[]' to 'string2[]'. If     *
* they match 0 is returned; if string1 < string2, a negative number     *
* giving the location of the first mismatch is returned, and if string1 *
* > string2, a positive number giving the mismatch location is returned *
\***********************************************************************/
int sstrcmp(char *string1, char *string2, int len)
{
    int i=0;

    while (i < len)
        if (string1[i] == string2[i])
            i++;
        else if (((unsigned char)(string1[i])) > ((unsigned char)(string2[i])))
            return (i+1);
        else
            return(-(i+1));
    return(0);
}
/*--------------------------END-OF-NLEDMISC.C-------------------------*/
